//
//	Gordon Wetzstein [gordon.wetzstein@medien.uni-weimar.de]
//	Faculty of Media
//	Bauhaus-University Weimar
//
//-----------------------------------------------------------------------------
// includes

#include <stdio.h>
#include <iostream>
using namespace std;

#ifdef WIN32
#include <windows.h>
#endif

//#define GLH_EXT_SINGLE_FILE
//#include <glh/glh_extensions.h>
#include <gl/glew.h>

#include "gl/glut.h"

#include "FramebufferObject.h"
FramebufferObject	_framebufferObject("rgba=4x8t depth=24t stencil=t");

static const char *modeTestStrings[] = 
{	
	"rgb=8t depth=32t",
	"rgba=8t depth=24t",
	"rgba=16t depth=24t",
	"rgba=32t depth=24t",
	"rgba=8t",
	"rgba=16t",
	"rgba=32t",
	"rgba=4x8t depth=24t stencil=t",
	"rgba=4x16t depth=24t stencil=t",
	"rgba=4x32t depth=24t stencil=t"
};

bool bDrawMRT = true;

static int numModeTestStrings = sizeof(modeTestStrings) / sizeof(char*);
static int currentString      = 0;


//-----------------------------------------------------------------------------

GLint	_frameBufferSize[]	= { 1024, 768 };	
bool	_bFBOInitialized	= false;

GLint _windowSize[] = { 640, 480 };

GLenum my_program;

//-----------------------------------------------------------------------------
// GLUT callbacks and functions

void initGlut(int argc, char **argv);
void displayFunc(void);
void idleFunc(void);
void reshapeFunc(int width, int height);
void mouseFunc(int button, int state, int x, int y);
void mouseMotionFunc(int x, int y);
void keyboardFunc(unsigned char key, int x, int y);
void specialFunc(int key, int x, int y);

//-----------------------------------------------------------------------------

// other [OpenGL] functions
void countFrames(void);
void renderBitmapString(float x, float y, float z, void *font, char *string);
void drawFont(void);
void drawFont2(void);

//-----------------------------------------------------------------------------

bool bFullsreen = false;
int nWindowID;

// camera attributes
float viewerPosition[3]			= { 0.0, 0.0, -50.0 };
float viewerDirection[3]		= { 0.0, 0.0, 0.0 };
float viewerUp[3]						= { 0.0, 1.0, 0.0 };

// rotation values for the navigation
float navigationRotation[3]	= { 0.0, 0.0, 0.0 };

//-----------------------------------------------------------------------------

// parameters for the framecounter
char pixelstring[30];
int cframe = 0;
int time = 0;
int timebase = 0;

//-----------------------------------------------------------------------------

// parameters for the navigation

// position of the mouse when pressed
int mousePressedX = 0, mousePressedY = 0;
float lastXOffset = 0.0, lastYOffset = 0.0, lastZOffset = 0.0;
// mouse button states
int leftMouseButtonActive = 0, middleMouseButtonActive = 0, rightMouseButtonActive = 0;
// modifier state
int shiftActive = 0, altActive = 0, ctrlActive = 0;

//-----------------------------------------------------------------------------

bool init = false;

//-----------------------------------------------------------------------------
//	shader stuff

bool	bShaderInitialized = false;
void	initShader(void);
void	drawContent(void);
void	drawMRTContent(void);

// destination color buffers
const GLenum buffers[] = {
	GL_COLOR_ATTACHMENT0_EXT,
	GL_COLOR_ATTACHMENT1_EXT,
	GL_COLOR_ATTACHMENT2_EXT,
	GL_COLOR_ATTACHMENT3_EXT
};

//-----------------------------------------------------------------------------

void displayFunc(void) {

	GLenum err;

	if(!_bFBOInitialized) {
		_framebufferObject.initialize( _frameBufferSize[0], _frameBufferSize[1] );
		_bFBOInitialized = true;

		err = glGetError();
		if(err != GL_NO_ERROR)
		printf("OpenGL ERROR after FBO initialization: %s\n", gluErrorString(err));
	}

	// -----------------------------------------------------------------------

	if(bDrawMRT) {
		drawMRTContent();
	} else {
		drawContent();
	}

	countFrames();
	
	// check for errors
	//GLenum err = glGetError();
	err = glGetError();
	if(err != GL_NO_ERROR)
		printf("OpenGL ERROR: %s\n", gluErrorString(err));

	glutSwapBuffers();
}

//-----------------------------------------------------------------------------

void	drawContent(void) {

	// -----------------------------------------------------------------------
	// -----------------------------------------------------------------------
	//	1. render teapot into color and depth buffer

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	_framebufferObject.beginCapture();

	// clear the buffers
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	_framebufferObject.disableTextureTarget();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, 1.33, 15.0, 70.0);
	
	glMatrixMode(GL_MODELVIEW);	
	glLoadIdentity();

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	GLfloat	lightpos[4] = { 5.0, 15.0, 10.0, 1.0 }; 
	glLightfv(GL_LIGHT0, GL_POSITION, lightpos);

	glTranslatef( viewerPosition[0], viewerPosition[1], viewerPosition[2] );
	// add navigation rotation

	glRotatef( navigationRotation[0], 1.0f, 0.0f, 0.0f );
	glRotatef( navigationRotation[1], 0.0f, 1.0f, 0.0f );

	glutSolidTeapot(10.0);

  _framebufferObject.endCapture();

	glPopAttrib();

	// -----------------------------------------------------------------------
	// -----------------------------------------------------------------------

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	// draw color buffer 
	_framebufferObject.bind();
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.25);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(0.5, 0.25);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(0.5, 0.75);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.0, 0.75);
	glEnd();

	// draw depth buffer 
	_framebufferObject.bindDepth();
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.5, 0.25);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(1.0, 0.25);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(1.0, 0.75);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.5, 0.75);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	drawFont2();
}

//-----------------------------------------------------------------------------

void	drawMRTContent(void) {

	if(!bShaderInitialized) {
		initShader();
		bShaderInitialized = true;
	}

	// -----------------------------------------------------------------------
	// -----------------------------------------------------------------------
	//	1. render teapot into color, depth and stencil buffer

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	_framebufferObject.beginCapture();

	// enable shader 
	glUseProgram(my_program);

	glDrawBuffers(sizeof(buffers) / sizeof(GLenum), buffers);

	// clear the buffers
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClearStencil(0x0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS, 0x1, 0x1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	_framebufferObject.disableTextureTarget();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, 1.33, 15.0, 70.0);
	
	glMatrixMode(GL_MODELVIEW);	
	glLoadIdentity();

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	GLfloat	lightpos[4] = { 5.0, 15.0, 10.0, 1.0 }; 
	glLightfv(GL_LIGHT0, GL_POSITION, lightpos);

	glTranslatef( viewerPosition[0], viewerPosition[1], viewerPosition[2] );
	// add navigation rotation

	glRotatef( navigationRotation[0], 1.0f, 0.0f, 0.0f );
	glRotatef( navigationRotation[1], 0.0f, 1.0f, 0.0f );

	glutSolidTeapot(10.0);

	glUseProgram(NULL);
  
	_framebufferObject.endCapture();

	glPopAttrib();

	// -----------------------------------------------------------------------
	// -----------------------------------------------------------------------
	//	2. render red plane only where stencil is 1

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	_framebufferObject.beginCapture();

	glDrawBuffer(GL_COLOR_ATTACHMENT3_EXT);

	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_EQUAL, 0x1, 0x1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	_framebufferObject.disableTextureTarget();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	glColor3f(1.0, 0.0, 0.0);

	glBegin(GL_QUADS);
		glVertex2f(0.0, 0.0);
		glVertex2f(1.0, 0.0);
		glVertex2f(1.0, 1.0);
		glVertex2f(0.0, 1.0);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glStencilFunc(GL_EQUAL, 0x0, 0x1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	_framebufferObject.disableTextureTarget();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	glColor3f(0.0, 0.0, 1.0);

	glBegin(GL_QUADS);
		glVertex2f(0.0, 0.0);
		glVertex2f(1.0, 0.0);
		glVertex2f(1.0, 1.0);
		glVertex2f(0.0, 1.0);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
  
	_framebufferObject.endCapture();

	glPopAttrib();

	// -----------------------------------------------------------------------
	// -----------------------------------------------------------------------

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);	
	glPushMatrix();
	glLoadIdentity();

	// draw color buffer 0
	_framebufferObject.bind(0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.5);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(0.333, 0.5);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(0.333, 1.0);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.0, 1.0);
	glEnd();

	// draw color buffer 1
	_framebufferObject.bind(1);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.0);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(0.333, 0.0);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(0.333, 0.5);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.0, 0.5);
	glEnd();

	// draw color buffer 2
	_framebufferObject.bind(2);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.333, 0.5);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(0.666, 0.5);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(0.666, 1.0);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.333, 1.0);
	glEnd();

	// draw color buffer 3
	_framebufferObject.bind(3);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.333, 0.0);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(0.666, 0.0);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(0.666, 0.5);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.333, 0.5);
	glEnd();

	// draw depth buffer 
	_framebufferObject.bindDepth();
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.666, 0.5);
		glTexCoord2f(_framebufferObject.getWidth(), 0.0);
		glVertex2f(1.0, 0.5);
		glTexCoord2f(_framebufferObject.getWidth(), _framebufferObject.getHeight());
		glVertex2f(1.0, 1.0);
		glTexCoord2f(0.0, _framebufferObject.getHeight());
		glVertex2f(0.666, 1.0);
	glEnd();

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

  	drawFont();
}

//-----------------------------------------------------------------------------

void initGlut(int argc, char **argv) {

	// GLUT Window Initialization:
	glutInit (&argc, argv);
	glutInitWindowSize (_windowSize[0], _windowSize[1]);
	glutInitWindowPosition(200, 100);
	glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
	nWindowID = glutCreateWindow ("TestFramebufferObject");

	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
		std::cout << "ERROR: could not initialize GLEW!" << std::endl;
	}

	// Register callbacks:
	glutDisplayFunc		(displayFunc);
	glutReshapeFunc		(reshapeFunc);
	glutKeyboardFunc	(keyboardFunc);
	glutSpecialFunc		(specialFunc);
	glutMouseFunc		(mouseFunc);
	glutMotionFunc		(mouseMotionFunc);
	glutIdleFunc		(idleFunc);

}

//-----------------------------------------------------------------------------

void	initShader(void) {

	const GLchar *my_fragment_shader_source =
		"void main(void){\n"
		"gl_FragData[0] = gl_Color;\n"
		"gl_FragData[1] = gl_TexCoord[0];\n"
		"gl_FragData[2] = vec4(gl_FragCoord.x/";
		
	string shaderAsString(my_fragment_shader_source);
	char widthAsString[64];
	sprintf(widthAsString, "%d", _frameBufferSize[0]);
	shaderAsString.append(widthAsString);
	shaderAsString.append(", gl_FragCoord.y/");
	char heightAsString[64];
	sprintf(heightAsString, "%d", _frameBufferSize[1]);
	shaderAsString.append(heightAsString);
	shaderAsString.append(", gl_FragCoord.z ,1.0);\n}\n");	

	GLenum my_fragment_shader;

	// Create Shader And Program Objects
	my_program = glCreateProgram();
	my_fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);

	const GLchar *myShader = (const GLchar*)(shaderAsString.c_str());
	// Load Shader Sources
	glShaderSource(my_fragment_shader, 1, &myShader , NULL);

	// Compile The Shaders
	glCompileShader(my_fragment_shader);

	// Attach The Shader Objects To The Program Object
	glAttachShader(my_program, my_fragment_shader);

	// Link The Program Object
	glLinkProgram(my_program);

	GLenum error = glGetError();
	if(error != GL_NO_ERROR)
		cout << "GL ERROR: " << gluErrorString(error) << endl;
}

//-----------------------------------------------------------------------------

void idleFunc(void) {
	glutPostRedisplay();
}

//-----------------------------------------------------------------------------

void reshapeFunc(int width, int height) {

	_windowSize[0] = width;
	_windowSize[1] = height;

	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(75.0, 1.33, 1.0, 1000.0);

	glMatrixMode(GL_MODELVIEW);

}

//-----------------------------------------------------------------------------


// mouse callback
void mouseFunc(int button, int state, int x, int y) {
	
	// get the modifiers
	switch (glutGetModifiers()) {

		case GLUT_ACTIVE_SHIFT:
			shiftActive = 1;
			break;
		case GLUT_ACTIVE_ALT:
			altActive	= 1;
			break;
		case GLUT_ACTIVE_CTRL:
			ctrlActive	= 1;
			break;
		default:
			shiftActive = 0;
			altActive	= 0;
			ctrlActive	= 0;
			break;
	}

	// get the mouse buttons
	if (button == GLUT_LEFT_BUTTON)
		if (state == GLUT_DOWN) {
			leftMouseButtonActive += 1;
		} else
			leftMouseButtonActive -= 1;
	else if (button == GLUT_MIDDLE_BUTTON)
		if (state == GLUT_DOWN) {
			middleMouseButtonActive += 1;
			lastXOffset = 0.0;
			lastYOffset = 0.0;
		} else
			middleMouseButtonActive -= 1;
	else if (button == GLUT_RIGHT_BUTTON)
		if (state == GLUT_DOWN) {
			rightMouseButtonActive += 1;
			lastZOffset = 0.0;
		} else
			rightMouseButtonActive -= 1;

//	if (altActive) {
		mousePressedX = x;
		mousePressedY = y;
//	}
}

//-----------------------------------------------------------------------------

void mouseMotionFunc(int x, int y) {
	
	float xOffset = 0.0, yOffset = 0.0, zOffset = 0.0;

	// navigation
//	if (altActive) {
	
		// rotatation
		if (leftMouseButtonActive) {

			navigationRotation[0] += ((mousePressedY - y) * 180.0f) / 200.0f;
			navigationRotation[1] += ((mousePressedX - x) * 180.0f) / 200.0f;

			mousePressedY = y;
			mousePressedX = x;

		}
		// panning
		else if (middleMouseButtonActive) {

			xOffset = (mousePressedX + x);
			if (!lastXOffset == 0.0) {
				viewerPosition[0]	-= (xOffset - lastXOffset) / 8.0;
				viewerDirection[0]	-= (xOffset - lastXOffset) / 8.0;
			}
			lastXOffset = xOffset;

			yOffset = (mousePressedY + y);
			if (!lastYOffset == 0.0) {
				viewerPosition[1]	+= (yOffset - lastYOffset) / 8.0;
				viewerDirection[1]	+= (yOffset - lastYOffset) / 8.0;	
			}	
			lastYOffset = yOffset;

		}
		// depth movement
		else if (rightMouseButtonActive) {
			zOffset = (mousePressedX + x);
			if (!lastZOffset == 0.0) {
				viewerPosition[2] -= (zOffset - lastZOffset) / 5.0;
				viewerDirection[2] -= (zOffset - lastZOffset) / 5.0;
			}
			lastZOffset = zOffset;
		}
//	}

}

//-----------------------------------------------------------------------------

void keyboardFunc(unsigned char key, int x, int y) {
    
	switch(key) {
		
		// -----------------------------------------

#ifdef WIN32
		// exit on escape
		case '\033':
			exit(0);
			break;
#endif

		// -----------------------------------------
			
		// switch to fullscreen
		case 'f':

			bFullsreen = !bFullsreen;
			if(bFullsreen) 
				glutFullScreen();
			else {
				glutSetWindow(nWindowID);
				glutPositionWindow(100, 100);
				glutReshapeWindow(640, 480);
			}
			break;

		case ' ':
			++currentString%=numModeTestStrings;

			if(	(currentString == numModeTestStrings-1) ||
				(currentString == numModeTestStrings-2) ||
				(currentString == numModeTestStrings-3) )
				bDrawMRT = true;
			else
				bDrawMRT = false;

			std::cout << "reinitialize Framebuffer Object to " << modeTestStrings[currentString] << std::endl;
			_framebufferObject.reInitialize(1024, 768, modeTestStrings[currentString]);
			break;
	}
}

//-----------------------------------------------------------------------------

void specialFunc(int key, int x, int y) {
	//printf("key pressed: %d\n", key);
}

//-----------------------------------------------------------------------------

void drawFont(void) {

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_RECTANGLE_ARB);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 150, 0, 100);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor3f(0.0, 1.0, 0.0);

	std::string string1("Color Attachment 1 - Shaded Teapot");  	
	renderBitmapString(4.5,55,0.0,GLUT_BITMAP_HELVETICA_10, (char*)(string1.c_str()));

	std::string string2("Color Attachment 2 - Texture Coordinates");  	
	renderBitmapString(2,5,0.0,GLUT_BITMAP_HELVETICA_10,(char*)(string2.c_str()));

	std::string string3("Color Attachment 3 - Fragment Coordinates");  	
	renderBitmapString(50.2,55,0.0,GLUT_BITMAP_HELVETICA_10,(char*)(string3.c_str()));

	std::string string4("Color Attachment 4 - two planes [red and");  	
	renderBitmapString(51.8,5,0.0,GLUT_BITMAP_HELVETICA_10,(char*)(string4.c_str()));
	std::string string4_1("blue] rendered using the Stencil Attachment");
	renderBitmapString(52,2,0.0,GLUT_BITMAP_HELVETICA_10,(char*)(string4_1.c_str()));

	std::string string5("Depth Attachment");  	
	renderBitmapString(115.5,55,0.0,GLUT_BITMAP_HELVETICA_10,(char*)(string5.c_str()));
		
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glPopAttrib();

}

//-----------------------------------------------------------------------------

void drawFont2(void) {

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_RECTANGLE_ARB);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 100, 0, 100);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor3f(0.0, 1.0, 0.0);

	std::string string1("Color Attachment - Shaded Teapot");  	
	renderBitmapString(12,20,0.0,GLUT_BITMAP_HELVETICA_10, (char*)(string1.c_str()));

	std::string string5("Depth Attachment");  	
	renderBitmapString(70,20,0.0,GLUT_BITMAP_HELVETICA_10,(char*)(string5.c_str()));
		
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glPopAttrib();

}

//-----------------------------------------------------------------------------

void countFrames(void)  {

	time=glutGet(GLUT_ELAPSED_TIME);
	cframe++;
	if (time - timebase > 50) {
		sprintf(pixelstring, "fps: %4.2f", cframe*1000.0/(time-timebase));		
		timebase = time;
		cframe = 0;
	// Draw status text and uni-logo:
	} 

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_RECTANGLE_ARB);
	glColor3f(0.0,1.0,0.0);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 200, 0, 200);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// render the string
	renderBitmapString(180,5,0.0,GLUT_BITMAP_HELVETICA_10,pixelstring);
		
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

//-----------------------------------------------------------------------------

void renderBitmapString(float x, float y, float z, void *font, char *string) {
  char *c;
  glRasterPos3f(x, y,z);
  for (c=string; *c != '\0'; c++) {
    glutBitmapCharacter(font, *c);
  }
}

//-----------------------------------------------------------------------------


void main(int argc, char **argv) {
	
	printf("TestFramebufferObject\n\tGordon Wetzstein [gordon.wetzstein@medien.uni-weimar.de]\n\n");
	printf("keys:\n\tf\t- toggle fullscreen\n\tesc\t- exit\n\n");
	printf("mouse:\n\tleft button\t- rotation\n\tmiddle button\t- panning\n\tright button\t- zoom in and out\n\n\n");

	printf("space\t- reinitialize FBO with different modes!\n\n\n");

	initGlut(argc, argv);

	printf("OpenGL version: %s\n", glGetString(GL_VERSION));
	printf("OpenGL vendor: %s\n", glGetString(GL_VENDOR));
	printf("OpenGL renderer: %s\n", glGetString(GL_RENDERER));
	
	glutMainLoop();
		
}



